unit LoggingProfiler;

//TODO: Remoting routines

{.$define LOG_ALL_EVENTS}
{.$define JIT_INLINE_AT_WILL}
{.$define JIT_USE_NGEN_VERSIONS}
{.$define GC_EVENTS_LOG_OBJ_REFS}

interface

uses
  Windows, ActiveX, Classes, ComObj, BaseProfiler, CorProf;

type
  TLoggingProfiler = class(TBaseProfiler, ICorProfilerCallback)
  protected
    // STARTUP/SHUTDOWN EVENTS
    procedure OnInitialize; override;
    procedure OnShutDown; override;
    // COR_PRF_MONITOR_APPDOMAIN_LOADS
    procedure AppDomainCreationStarted(appDomainId: ULONG); safecall;
    procedure AppDomainCreationFinished(appDomainId: ULONG; hrStatus: HResult); safecall;
    procedure AppDomainShutdownStarted(appDomainId: ULONG); safecall;
    procedure AppDomainShutdownFinished(appDomainId: ULONG; hrStatus: HResult); safecall;
    // COR_PRF_MONITOR_ASSEMBLY_LOADS
    procedure AssemblyLoadStarted(assemblyId: ULONG); safecall;
    procedure AssemblyLoadFinished(assemblyId: ULONG; hrStatus: HResult); safecall;
    procedure AssemblyUnloadStarted(assemblyId: ULONG); safecall;
    procedure AssemblyUnloadFinished(assemblyId: ULONG; hrStatus: HResult); safecall;
    // COR_PRF_MONITOR_MODULE_LOADS
    procedure ModuleLoadStarted(moduleId: ULONG); safecall;
    procedure ModuleLoadFinished(moduleId: ULONG; hrStatus: HResult); safecall;
    procedure ModuleUnloadStarted(moduleId: ULONG); safecall;
    procedure ModuleUnloadFinished(moduleId: ULONG; hrStatus: HResult); safecall;
    procedure ModuleAttachedToAssembly(moduleId: ULONG; assemblyId: ULONG); safecall;
    // COR_PRF_MONITOR_CLASS_LOADS
    procedure ClassLoadStarted(classId: ULONG); safecall;
    procedure ClassLoadFinished(classId: ULONG; hrStatus: HResult); safecall;
    procedure ClassUnloadStarted(classId: ULONG); safecall;
    procedure ClassUnloadFinished(classId: ULONG; hrStatus: HResult); safecall;
    // COR_PRF_MONITOR_JIT_COMPILATION
    procedure JITCompilationStarted(functionId: ULONG; fIsSafeToBlock: Integer); safecall;
    procedure JITCompilationFinished(functionId: ULONG; hrStatus: HResult;
      fIsSafeToBlock: Integer); safecall;
    function JITInlining(callerId: ULONG; calleeId: ULONG): {fShouldInline} Bool; safecall;
    // COR_PRF_MONITOR_FUNCTION_UNLOADS
    procedure JITFunctionPitched(functionId: ULONG); safecall;
    // COR_PRF_MONITOR_CACHE_SEARCHES
    function JITCachedFunctionSearchStarted(
      functionId: ULONG): {bUseCachedFunction} Bool; safecall;
{*} procedure JITCachedFunctionSearchFinished(functionId: ULONG;
      AResult: COR_PRF_JIT_CACHE); safecall; //doesn't seem to fire
    // COR_PRF_MONITOR_THREADS
    procedure ThreadCreated(threadId: ULONG); safecall;
    procedure ThreadDestroyed(threadId: ULONG); safecall;
    procedure ThreadAssignedToOSThread(managedThreadId: ULONG; osThreadId: ULONG); safecall;
    // COR_PRF_MONITOR_CODE_TRANSITIONS
    procedure UnmanagedToManagedTransition(functionId: ULONG;
      reason: COR_PRF_TRANSITION_REASON); safecall;
    procedure ManagedToUnmanagedTransition(functionId: ULONG;
      reason: COR_PRF_TRANSITION_REASON); safecall;
    // COR_PRF_MONITOR_SUSPENDS
    procedure RuntimeSuspendStarted(suspendReason: COR_PRF_SUSPEND_REASON); safecall;
    procedure RuntimeSuspendFinished; safecall;
    procedure RuntimeSuspendAborted; safecall;
    procedure RuntimeResumeStarted; safecall;
    procedure RuntimeResumeFinished; safecall;
    procedure RuntimeThreadSuspended(threadId: ULONG); safecall;
    procedure RuntimeThreadResumed(threadId: ULONG); safecall;
    // COR_PRF_MONITOR_GC
    procedure MovedReferences(cMovedObjectIDRanges: ULONG;
      var oldObjectIDRangeStart, newObjectIDRangeStart,
          cObjectIDRangeLength: TULONGArray); safecall;
    procedure ObjectsAllocatedByClass(cClassCount: ULONG;
      var classIds, cObjects: TULONGarray); safecall;
    procedure ObjectReferences(objectId: ULONG; classId: ULONG;
      cObjectRefs: ULONG; var objectRefIds: TULONGArray); safecall;
    procedure RootReferences(cRootRefs: ULONG; var rootRefIds: TULONGArray); safecall;
    // COR_PRF_MONITOR_OBJECT_ALLOCATED or COR_PRF_ENABLE_OBJECT_ALLOCATED
    procedure ObjectAllocated(objectId, classId: ULONG); safecall;
    // COR_PRF_MONITOR_CCW
    procedure COMClassicVTableCreated(wrappedClassId: ULONG;
      const implementedIID: TGuid; pVTable: Pointer; cSlots: ULONG); safecall;
    procedure COMClassicVTableDestroyed(wrappedClassId: ULONG;
      const implementedIID: TGuid; pVTable: Pointer); safecall;
  end;

const
  ProfilerCoClassName = 'LoggingProfiler';
  ProfilerDescription = 'Logging profiler';
  ProfilerGUID: TGUID = '{0794BEB3-0C61-4B1D-B4B0-EFA4A0139206}';
  ProfilerClass: TComClass = TLoggingProfiler;

implementation

uses
  ComServ, SysUtils;

{ TLoggingProfiler }

procedure TLoggingProfiler.OnInitialize;
begin
  NestingLevel := 0;
{$ifdef LOG_ALL_EVENTS}
  EventMask := COR_PRF_MONITOR_ALL;
{$else}
  EventMask := COR_PRF_ENABLE_REJIT;
  {$ifndef JIT_INLINE_AT_WILL}
  EventMask := EventMask or COR_PRF_DISABLE_INLINING;
  {$endif}
  EventMask := EventMask or COR_PRF_MONITOR_FUNCTION_UNLOADS;
  EventMask := EventMask or COR_PRF_MONITOR_CLASS_LOADS;
  EventMask := EventMask or COR_PRF_MONITOR_MODULE_LOADS;
  EventMask := EventMask or COR_PRF_MONITOR_ASSEMBLY_LOADS;
  EventMask := EventMask or COR_PRF_MONITOR_APPDOMAIN_LOADS;
  EventMask := EventMask or COR_PRF_MONITOR_JIT_COMPILATION;
  EventMask := EventMask or COR_PRF_MONITOR_EXCEPTIONS;
  EventMask := EventMask or COR_PRF_MONITOR_GC;
  EventMask := EventMask or COR_PRF_ENABLE_OBJECT_ALLOCATED;
  EventMask := EventMask or COR_PRF_MONITOR_OBJECT_ALLOCATED;
  EventMask := EventMask or COR_PRF_MONITOR_THREADS;
  EventMask := EventMask or COR_PRF_MONITOR_CODE_TRANSITIONS;
  EventMask := EventMask or COR_PRF_MONITOR_ENTERLEAVE;
  EventMask := EventMask or COR_PRF_MONITOR_CCW;
  EventMask := EventMask or COR_PRF_MONITOR_SUSPENDS;
  EventMask := EventMask or COR_PRF_MONITOR_CACHE_SEARCHES;
  EventMask := EventMask or COR_PRF_MONITOR_CLR_EXCEPTIONS;
{$endif}
  Log('CLR has started')
end;

procedure TLoggingProfiler.OnShutDown;
begin
  Log('CLR is shutting down');
end;

procedure TLoggingProfiler.AppDomainCreationStarted(appDomainId: ULONG);
var
  AppDomainName: WideString;
begin
  SyncEnter;
  if GetAppDomainFromAppDomainId(appDomainId, AppDomainName) then
    Log('AppDomain creation started,  name: %s', [AppDomainName])
  else
    Log('AppDomain creation started,  ID: $%x', [appDomainID]);
  Inc(NestingLevel);
  SyncExit  
end;

procedure TLoggingProfiler.AppDomainCreationFinished(appDomainId: ULONG;
  hrStatus: HResult);
var
  AppDomainName: WideString;
begin
  SyncEnter;
  Dec(NestingLevel);
  if GetAppDomainFromAppDomainId(appDomainId, AppDomainName) then
    Log('AppDomain creation finished, name: %s, status: $%x',
      [AppDomainName, hrStatus])
  else
    Log('AppDomain creation finished, ID: $%x, status: $%x',
      [appDomainID, hrStatus]);
  SyncExit      
end;

procedure TLoggingProfiler.AppDomainShutdownStarted(appDomainId: ULONG);
var
  AppDomainName: WideString;
begin
  SyncEnter;
  if GetAppDomainFromAppDomainId(appDomainId, AppDomainName) then
    Log('AppDomain shutting down,  name: %s', [AppDomainName])
  else
    Log('AppDomain shutting down,  ID: $%x', [appDomainID]);
  Inc(NestingLevel);
  SyncExit
end;

procedure TLoggingProfiler.AppDomainShutdownFinished(appDomainId: ULONG;
  hrStatus: HResult);
var
  AppDomainName: WideString;
begin
  SyncEnter;
  Dec(NestingLevel);
  if GetAppDomainFromAppDomainId(appDomainId, AppDomainName) then
    Log('AppDomain shutdown finished, name: %s', [AppDomainName])
  else
    Log('AppDomain shutdown finished, ID: $%x', [appDomainID]);
  SyncExit
end;

procedure TLoggingProfiler.AssemblyLoadStarted(assemblyId: ULONG);
var
  AssemblyName: WideString;
begin
  SyncEnter;
  if GetAssemblyFromAssemblyId(assemblyId, AssemblyName) then
    Log('Assembly load started,  name: %s', [AssemblyName])
  else
    Log('Assembly load started,  ID: $%x', [assemblyID]);
  Inc(NestingLevel);
  SyncExit  
end;

procedure TLoggingProfiler.AssemblyLoadFinished(assemblyId: ULONG;
  hrStatus: HResult);
var
  AssemblyName: WideString;
begin
  SyncEnter;
  Dec(NestingLevel);
  if GetAssemblyFromAssemblyId(assemblyId, AssemblyName) then
    Log('Assembly load finished, name: %s, status: $%x', [AssemblyName, hrStatus])
  else
    Log('Assembly load finished, ID: $%x, status: $%x', [assemblyID, hrStatus]);
  SyncExit    
end;

procedure TLoggingProfiler.AssemblyUnloadStarted(assemblyId: ULONG);
var
  AssemblyName: WideString;
begin
  SyncEnter;
  if GetAssemblyFromAssemblyId(assemblyId, AssemblyName) then
    Log('Assembly unload started,  name: %s', [AssemblyName])
  else
    Log('Assembly unload started,  ID: $%x', [assemblyId]);
  Inc(NestingLevel);
  SyncExit
end;

procedure TLoggingProfiler.AssemblyUnloadFinished(assemblyId: ULONG;
  hrStatus: HResult);
var
  AssemblyName: WideString;
begin
  SyncEnter;
  Dec(NestingLevel);
  if GetAssemblyFromAssemblyId(assemblyId, AssemblyName) then
    Log('Assembly unload finished, name: %s', [AssemblyName])
  else
    Log('Assembly unload finished, ID: $%x', [assemblyId]);
  SyncExit    
end;

procedure TLoggingProfiler.ModuleLoadStarted(moduleId: ULONG);
var
  ModuleName: WideString;
begin
  SyncEnter;
  if GetModuleFromModuleId(moduleId, ModuleName) then
    Log('Module load started,  name: %s', [ModuleName])
  else
    Log('Module load started,  ID: $%x', [moduleID]);
  Inc(NestingLevel);
  SyncExit  
end;

procedure TLoggingProfiler.ModuleLoadFinished(moduleId: ULONG;
  hrStatus: HResult);
var
  ModuleName: WideString;
begin
  SyncEnter;
  Dec(NestingLevel);
  if GetModuleFromModuleId(moduleId, ModuleName) then
    Log('Module load finished, name: %s, status $%x',
      [ModuleName, hrStatus])
  else
    Log('Module load finished, ID: $%x, status $%x',
      [moduleID, hrStatus]);
  SyncExit      
end;

procedure TLoggingProfiler.ModuleUnloadStarted(moduleId: ULONG);
var
  ModuleName: WideString;
begin
  SyncEnter;
  if GetModuleFromModuleId(moduleId, ModuleName) then
    Log('Module unload started,  name: %s', [ModuleName])
  else
    Log('Module unload started,  ID: %s', [moduleID]);
  Inc(NestingLevel);
  SyncExit
end;

procedure TLoggingProfiler.ModuleUnloadFinished(moduleId: ULONG;
  hrStatus: HResult);
var
  ModuleName: WideString;
begin
  SyncEnter;
  Dec(NestingLevel);
  if GetModuleFromModuleId(moduleId, ModuleName) then
    Log('Module unload finished, name: %s, status: $%x', [ModuleName, hrStatus])
  else
    Log('Module unload finished, ID: $%x, status: $%x', [moduleID, hrStatus]);
  SyncExit    
end;

procedure TLoggingProfiler.ModuleAttachedToAssembly(moduleId,
  assemblyId: ULONG);
var
  ModuleName, AssemblyName: WideString;
begin
  SyncEnter;
  if GetModuleFromModuleId(moduleId, ModuleName) and
     GetAssemblyFromAssemblyId(assemblyId, AssemblyName) then
    Log('Module %s attached to assembly %s', [ModuleName, AssemblyName])
  else
    Log('Module (ID: $%x) attached to assembly (ID: $%x)',
      [moduleID, assemblyID]);
  SyncExit      
end;

procedure TLoggingProfiler.ClassLoadStarted(classId: ULONG);
var
  ClassName: WideString;
begin
  SyncEnter;
  if GetClassFromClassId(classId, ClassName) then
    Log('Class load started,  name: %s', [ClassName])
  else
    Log('Class load started,  ID: $%x', [classID]);
  Inc(NestingLevel);
  SyncExit  
end;

procedure TLoggingProfiler.ClassLoadFinished(classId: ULONG;
  hrStatus: HResult);
var
  ClassName: WideString;
begin
  SyncEnter;
  Dec(NestingLevel);
  if GetClassFromClassId(classId, ClassName) then
    Log('Class load finished, name: %s, status $%x', [ClassName, hrStatus])
  else
    Log('Class load finished, ID: $%x, status $%x', [classID, hrStatus]);
  SyncExit
end;

procedure TLoggingProfiler.ClassUnloadStarted(classId: ULONG);
var
  ClassName: WideString;
begin
  SyncEnter;
  if GetClassFromClassId(classId, ClassName) then
    Log('Class unload started,  name: %s', [ClassName])
  else
    Log('Class unload started,  ID: $%x', [classID]);
  Inc(NestingLevel);
  SyncExit  
end;

procedure TLoggingProfiler.ClassUnloadFinished(classId: ULONG;
  hrStatus: HResult);
var
  ClassName: WideString;
begin
  SyncEnter;
  Dec(NestingLevel);
  if GetClassFromClassId(classId, ClassName) then
    Log('Class unload finished, name: %s', [ClassName])
  else
    Log('Class unload finished, ID: $%x', [classID]);
  SyncExit
end;

procedure TLoggingProfiler.JITCompilationStarted(functionId: ULONG;
  fIsSafeToBlock: Integer);
var
  ClassName, MethodName: WideString;
begin
  SyncEnter;
  if GetClassAndMethodFromFunctionId(functionID, ClassName, MethodName) then
    Log('JIT compilation started,  name: %s.%s', [ClassName, MethodName])
  else
    Log('JIT compilation started,  ID: $%x', [functionID]);
  Inc(NestingLevel);
  SyncExit
end;

procedure TLoggingProfiler.JITCompilationFinished(functionId: ULONG;
  hrStatus: HResult; fIsSafeToBlock: Integer);
var
  ClassName, MethodName: WideString;
begin
  SyncEnter;
  Dec(NestingLevel);
  if GetClassAndMethodFromFunctionId(functionID, ClassName, MethodName) then
    Log('JIT compilation finished, name: %s.%s, status: $%x',
      [ClassName, MethodName, hrStatus])
  else
    Log('JIT compilation finished, ID: $%x, status: $%x', [functionID, hrStatus]);
  SyncExit    
end;

procedure TLoggingProfiler.JITFunctionPitched(functionId: ULONG);
var
  ClassName, MethodName: WideString;
begin
  SyncEnter;
  if GetClassAndMethodFromFunctionId(functionID, ClassName, MethodName) then
    Log('Function pitched, name: %s.%s', [ClassName, MethodName])
  else
    Log('Function pitched, ID: $%x', [functionID]);
  SyncExit    
end;

function TLoggingProfiler.JITInlining(callerId, calleeId: ULONG): Bool;
var
  Class1Name, Method1Name, Class2Name, Method2Name: WideString;
begin
  SyncEnter;
  if GetClassAndMethodFromFunctionId(calleeId, Class1Name, Method1Name) and
     GetClassAndMethodFromFunctionId(callerId, Class2Name, Method2Name) then
    Log('JIT inlining %s.%s into %s.%s',
      [Class1Name, Method1Name, Class2Name, Method2Name])
  else
    Log('JIT inlining');
{$ifdef JIT_INLINE_AT_WILL}
  //Let JIT inline everything it wants to
  Result := True;
{$else}
  //Force JIT to inline NOTHING
  Result := False;
{$endif}
  SyncExit
end;

function TLoggingProfiler.JITCachedFunctionSearchStarted(
  functionId: ULONG): Bool;
var
  ClassName, MethodName: WideString;
begin
  SyncEnter;
  if GetClassAndMethodFromFunctionId(functionID, ClassName, MethodName) then
    Log('JIT cached fn. search started,  name: %s.%s',
      [ClassName, MethodName])
  else
    Log('JIT cached fn. search started,  ID: $%x', [functionID]);
{$ifdef JIT_USE_NGEN_VERSIONS}
  //Let JIT use cached version if possible
  Result := True;
{$else}
  //Force JITter to compile
  Result := False;
{$endif}
  //The JITCachedFunctionSearchFinished event is not fired, so best not indent here
  //Inc(NestingLevel);
  SyncExit
end;

//This event doesn't seem to fire, even
//though JITCachedFunctionSearchStarted does
procedure TLoggingProfiler.JITCachedFunctionSearchFinished(functionId: ULONG;
  AResult: COR_PRF_JIT_CACHE);
var
  ClassName, MethodName: WideString;
const
  Results: array[Boolean] of String = ('not found', 'found');
begin
  SyncEnter;
  //Dec(NestingLevel);
  if GetClassAndMethodFromFunctionId(functionID, ClassName, MethodName) then
    Log('JIT cached fn. search finished, name: %s.%s, cached version %s',
      [ClassName, MethodName, Results[AResult = COR_PRF_CACHED_FUNCTION_FOUND]])
  else
    Log('JIT cached fn. search finished, ID: $%x',
      [functionID, Results[AResult = COR_PRF_CACHED_FUNCTION_FOUND]]);
  SyncExit
end;

procedure TLoggingProfiler.ThreadCreated(threadId: ULONG);
begin
  SyncEnter;
  NestingLevel := 0;
  Log('Thread created, ID: $%x', [threadId]);
  SyncExit  
end;

procedure TLoggingProfiler.ThreadDestroyed(threadId: ULONG);
begin
  SyncEnter;
  Log('Thread destroyed, ID: $%x', [threadId]);
  SyncExit  
end;

procedure TLoggingProfiler.ThreadAssignedToOSThread(managedThreadId,
  osThreadId: ULONG);
begin
  SyncEnter;
  Log('Thread (ID: $%x) assigned to OS thread (ID: $%x)',
    [managedThreadId, osThreadId]);
  SyncExit
end;

procedure TLoggingProfiler.UnmanagedToManagedTransition(functionId: ULONG;
  reason: COR_PRF_TRANSITION_REASON);
var
  ClassName, MethodName: WideString;
begin
  SyncEnter;
  if reason = COR_PRF_TRANSITION_RETURN then
    Dec(NestingLevel);
  if GetClassAndMethodFromFunctionId(functionId, ClassName, MethodName) then
    Log('Unmanaged to managed transition %s.%s', [ClassName, MethodName])
  else
    Log('Unmanaged to managed transition');
  if reason = COR_PRF_TRANSITION_CALL then
    Inc(NestingLevel);
  SyncExit    
end;

procedure TLoggingProfiler.ManagedToUnmanagedTransition(functionId: ULONG;
  reason: COR_PRF_TRANSITION_REASON);
var
  ClassName, MethodName: WideString;
begin
  SyncEnter;
  if reason = COR_PRF_TRANSITION_RETURN then
    Dec(NestingLevel);
  if GetClassAndMethodFromFunctionId(functionId, ClassName, MethodName) then
    Log('Managed to unmanaged transition %s.%s', [ClassName, MethodName])
  else
    Log('Managed to unmanaged transition');
  if reason = COR_PRF_TRANSITION_CALL then
    Inc(NestingLevel);
  SyncExit
end;

procedure TLoggingProfiler.RuntimeSuspendStarted(
  suspendReason: COR_PRF_SUSPEND_REASON);
var
  Msg: String;
begin
  SyncEnter;
  Msg := 'CLR suspend started for ';
  case suspendReason of
    COR_PRF_SUSPEND_OTHER: Msg := Msg + 'some reason';
    COR_PRF_SUSPEND_FOR_GC: Msg := Msg + 'garbage collection';
    COR_PRF_SUSPEND_FOR_APPDOMAIN_SHUTDOWN: Msg := Msg + 'an AppDomain shutdown';
    //not implemented for the v1 JITter
    COR_PRF_SUSPEND_FOR_CODE_PITCHING: Msg := Msg + 'code pitching';
    COR_PRF_SUSPEND_FOR_SHUTDOWN: Msg := Msg + 'shutdown';
    COR_PRF_SUSPEND_FOR_INPROC_DEBUGGER: Msg := Msg + 'in-process debugging';
    COR_PRF_SUSPEND_FOR_GC_PREP: Msg := Msg + 'garbage collection preparation';
  end;
  Log(Msg);
  Inc(NestingLevel);
  SyncExit
end;

procedure TLoggingProfiler.RuntimeSuspendFinished;
begin
  SyncEnter;
  Dec(NestingLevel);
  Log('CLR suspend finished');
  SyncExit  
end;

procedure TLoggingProfiler.RuntimeSuspendAborted;
begin
  SyncEnter;
  Dec(NestingLevel);
  Log('CLR suspend aborted');
  SyncExit  
end;

procedure TLoggingProfiler.RuntimeResumeStarted;
begin
  SyncEnter;
  Log('CLR resume started');
  Inc(NestingLevel);
  SyncExit  
end;

procedure TLoggingProfiler.RuntimeResumeFinished;
begin
  SyncEnter;
  Dec(NestingLevel);
  Log('CLR resume finished');
  SyncExit  
end;

procedure TLoggingProfiler.RuntimeThreadSuspended(threadId: ULONG);
begin
  SyncEnter;
  Log('CLR thread $%d suspended', [threadID]);
  SyncExit
end;

procedure TLoggingProfiler.RuntimeThreadResumed(threadId: ULONG);
begin
  SyncEnter;
  Log('CLR thread $%d resumed', [threadID]);
  SyncExit  
end;

procedure TLoggingProfiler.MovedReferences(cMovedObjectIDRanges: ULONG;
  var oldObjectIDRangeStart, newObjectIDRangeStart,
      cObjectIDRangeLength: TULONGArray);
begin
  Log('Moved references');
end;

procedure TLoggingProfiler.ObjectAllocated(objectId, classId: ULONG);
var
  ClassName: WideString;
begin

  SyncEnter;
  if GetClassFromClassId(classId, ClassName) then
    Log('Object allocated of type %s', [ClassName])
  else
    Log('Object allocated, ID: $%x', [classID]);
  SyncExit
end;

procedure TLoggingProfiler.ObjectsAllocatedByClass(cClassCount: ULONG;
  var classIds, cObjects: TULONGarray);
begin
  SyncEnter;
  Log('Classes of objects in generation 0 = %d', [cClassCount]);
  SyncExit
end;

procedure TLoggingProfiler.ObjectReferences(objectId: ULONG; classId: ULONG;
  cObjectRefs: ULONG; var objectRefIds: TULONGArray);
{$ifdef GC_EVENTS_LOG_OBJ_REFS}
var
  ClassName: WideString;
{$endif}
begin
{$ifdef GC_EVENTS_LOG_OBJ_REFS}
  //This code gets called _loads_ each time a GC occurs as we get told
  //the references owend by _every_ object. We'll generally not bother
  //doing anything as it slows things down considerably
  SyncEnter;
  if GetClassFromClassId(classId, ClassName) then
    Log('Object references held by %s object, ID $%x = %d',
      [ClassName, objectId, cObjectRefs])
  else
    Log('Object references held by object, ID $%x = %d',
      [objectId, cObjectRefs]);
  SyncExit
{$endif}
end;

procedure TLoggingProfiler.RootReferences(cRootRefs: ULONG;
  var rootRefIds: TULONGArray);
begin
  SyncEnter;
  Log('Root references found = %d', [cRootRefs]);
  SyncExit
end;

procedure TLoggingProfiler.COMClassicVTableCreated(wrappedClassId: ULONG;
  const implementedIID: TGuid; pVTable: Pointer; cSlots: ULONG);
var
  ClassName: WideString;
  Guid: String;
begin
  SyncEnter;
  if IsEqualGUID(implementedIID, GUID_NULL) then
    Guid := ''
  else
    Guid := ' ' + GuidToString(implementedIID);
  if GetClassFromClassId(wrappedclassId, ClassName) then
    Log('CCW vtable created for %s%s at $%p with %d slots',
      [ClassName, Guid, pVTable, cSlots, GuidToString(implementedIID)])
  else
    Log('CCW vtable created, class ID: $%x%s, address: $%p, #slots: %d',
      [wrappedClassID, Guid, pVTable, cSlots, GuidToString(implementedIID)]);
  SyncExit      
end;

procedure TLoggingProfiler.COMClassicVTableDestroyed(wrappedClassId: ULONG;
  const implementedIID: TGuid; pVTable: Pointer);
var
  ClassName: WideString;
  Guid: String;
begin
  SyncEnter;
  if IsEqualGUID(implementedIID, GUID_NULL) then
    Guid := ''
  else
    Guid := ' ' + GuidToString(implementedIID);
  if GetClassFromClassId(wrappedclassId, ClassName) then
    Log('CCW vtable destroyed for %s%s at $%p',
      [ClassName, Guid, pVTable, GuidToString(implementedIID)])
  else
    Log('CCW vtable destroyed, class ID: $%x%s, address: $%p',
      [wrappedClassID, pVTable, GuidToString(implementedIID)]);
  SyncExit
end;

initialization
  TComObjectFactory.Create(ComServer, ProfilerClass, ProfilerGUID,
    ProfilerCoClassName, ProfilerDescription, ciMultiInstance, tmFree);
end.
