unit TraceProfiler;

//TODO: Add code to ignore common startup nonsense
//TODO: Add in code to ignore finalizer thread using function mapper
//TODO: Add in code to log each thread's trace separately

interface

uses
  Windows, ActiveX, Classes, ComObj, BaseProfiler, CorProf;

type
  TTraceProfiler = class(TBaseProfiler, ICorProfilerCallback)
  protected
    procedure OnInitialize; override;
    procedure FuncEnter(functionID: ULONG);
    procedure FuncLeave(functionID: ULONG);
    procedure FuncTailCall(functionID: ULONG);
  end;

const
  ProfilerCoClassName = 'TraceProfiler';
  ProfilerDescription = 'Trace profiler';
  ProfilerGUID: TGUID = '{0403B90F-6E5B-423C-A08D-760D2D851D2A}';
  ProfilerClass: TComClass = TTraceProfiler;

implementation

uses
  ComServ, SysUtils;

var
  Profiler: TTraceProfiler;

procedure NakedEnter(functionID: ULONG); assembler; stdcall;
asm
  push eax
  push ecx
  push edx
  mov edx, functionID
  mov eax, Profiler
  call TTraceProfiler.FuncEnter
  pop edx
  pop ecx
  pop eax
end;

procedure NakedLeave(functionID: ULONG); assembler; stdcall;
asm
  push eax
  push ecx
  push edx
  mov edx, functionID
  mov eax, Profiler
  call TTraceProfiler.FuncLeave
  pop edx
  pop ecx
  pop eax
end;

//Tailcall is an ununsed optimisation in JIT 1.x
procedure NakedTailCall(functionID: ULONG); assembler; stdcall;
asm
  push eax
  push ecx
  push edx
  mov edx, functionID
  mov eax, Profiler
  call TTraceProfiler.FuncTailCall
  pop edx
  pop ecx
  pop eax
end;

{ TTraceProfiler }

procedure TTraceProfiler.FuncEnter(functionID: ULONG);
var
  ClassName, MethodName: WideString;
begin
  SyncEnter;
  if GetClassAndMethodFromFunctionId(functionID, ClassName, MethodName) then
    Log('%s.%s entered', [ClassName, MethodName])
  else
    Log('Method id: $%x entered', [functionID]);
  Inc(NestingLevel);
  SyncExit;
end;

procedure TTraceProfiler.FuncLeave(functionID: ULONG);
var
  ClassName, MethodName: WideString;
begin
  SyncEnter;
  Dec(NestingLevel);
  if GetClassAndMethodFromFunctionId(functionID, ClassName, MethodName) then
    Log('%s.%s exited', [ClassName, MethodName])
  else
    Log('Method id: $%x exited', [functionID]);
  SyncExit;
end;

procedure TTraceProfiler.FuncTailCall(functionID: ULONG);
var
  ClassName, MethodName: WideString;
begin
  SyncEnter;
  Dec(NestingLevel);
  if GetClassAndMethodFromFunctionId(functionID, ClassName, MethodName) then
    Log('%s.%s exited (tailcall)', [ClassName, MethodName])
  else
    Log('Method id: $%x exited (tailcall)', [functionID]);
  SyncExit;
end;

procedure TTraceProfiler.OnInitialize;
begin
  EventMask := COR_PRF_MONITOR_ENTERLEAVE or COR_PRF_DISABLE_INLINING;
  //Store reference to profiler object, so external proc.s can call its methods
  Profiler := Self;
  CorProfilerInfo.SetEnterLeaveFunctionHooks(NakedEnter, NakedLeave, NakedTailCall);
end;

initialization
  TComObjectFactory.Create(ComServer, ProfilerClass, ProfilerGUID,
    ProfilerCoClassName, ProfilerDescription, ciMultiInstance, tmFree);
end.
